import asyncio
import importlib
import inspect
import logging
import traceback
from asyncio import Future, CancelledError
from typing import (
    TypeVar,
    Tuple,
    List,
    Optional,
    Union,
    Callable,
    AsyncIterator,
    Awaitable,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")
Input = TypeVar("Input", contravariant=True)
Output = TypeVar("Output", covariant=True)
AwaitableOrObj = Union[T, Awaitable[T]]

SourceConvertable = Optional[Union[Callable[[], T], T]]

OptionalMultipleOutputs = Union[AsyncIterator[T], Tuple]


class _MarkerObjectType:
    """This is a marker object used to indicate a special case during stream iteration."""

    pass


EndOfStreamMarker = _MarkerObjectType()
"""Special marker object that indicates the end of a stream.
"""


QueueExceptionMarker = _MarkerObjectType()
"""Special marker object that indicates an exception.
"""


def to_source(x: SourceConvertable) -> AsyncIterator[T]:
    """Creates a source from a callable or value.

    Tries to turn the input value into a VoiceStream source.
    - If it is a callable, assumes that is a callable that returns an AsyncIterator
    - If it is an object, return a single_source containing the object.
    - If it is None, return an empty_source.

    To create a source that explicitly yields `None`, pass lambda x: none_source()
    """
    if callable(x):
        # logger.debug("Callable source")
        return x()
    elif x is None:
        # logger.debug("Empty source")
        from voice_stream.core import empty_source

        return empty_source()
    else:
        # logger.debug(f"Single source {x}")
        from voice_stream.core import single_source

        return single_source(x)


def to_tuple(obj) -> Tuple:
    if isinstance(obj, Tuple):
        return obj
    if isinstance(obj, List):
        return tuple(obj)
    else:
        return (obj,)


def from_tuple(obj):
    return obj[0] if len(obj) == 1 else obj


def is_async_iterator(obj):
    return inspect.iscoroutinefunction(getattr(obj, "__anext__", None)) and hasattr(
        obj, "__aiter__"
    )


def map_future(f: Future[T], func: Callable[[T], Output]) -> Future[Output]:
    """Returns a future that will be resolved with the result of applying func to the result of f.

    Words are made lowercase and punctuation is removed
    before counting.

    Parameters
    ----------
    input_file : str
        Path to text file.

    Returns
    -------
    collections.Counter
        dict-like object where keys are words and values are counts.

    Examples
    --------
    >>> count_words("text.txt")
    """
    loop = asyncio.get_running_loop()
    out = loop.create_future()

    def callback(fut: Future[T]) -> None:
        try:
            out.set_result(func(fut.result()))
        except Exception as e:
            out.set_exception(e)

    f.add_done_callback(callback)
    return out


async def resolve_awaitable_or_obj(obj: AwaitableOrObj[T]) -> T:
    """Returns the result of an object or a future"""
    if inspect.isawaitable(obj):
        return await obj
    return obj


def load_attribute(full_path):
    # Split the path into module and attribute
    parts = full_path.split(".")
    module_path = ".".join(parts[:-1])
    attribute_name = parts[-1]

    try:
        # Import the module
        module = importlib.import_module(module_path)
        # Return the attribute
        return getattr(module, attribute_name)
    except ImportError:
        print(f"Module {module_path} not found.")
    except AttributeError:
        print(f"Attribute {attribute_name} not found in module {module_path}.")


async def cancel_with_confirmation(task: asyncio.Task) -> None:
    """Cancels a task and waits for confirmation that it has stopped."""
    task.cancel()
    try:
        a = await task
        # logger.debug(f"Return {a}")
    except asyncio.CancelledError:
        # logger.debug("C")
        pass
    except StopAsyncIteration:
        # logger.debug("S")
        pass
    # logger.debug(f"Cancelled task {format_task(task)}")


def format_task(task: asyncio.Task):
    if task.cancelled():
        if task.done():
            state = "canceled"
        else:
            state = "cancelling"
    elif task.done():
        result = "exception" if task.exception() else "normal"
        state = f"finished ({result})"
    else:
        state = "pending"  # The task is still pending or running
    return task.get_name() + " " + state


def format_current_task():
    return format_task(asyncio.current_task())


def format_stack_trace(task: asyncio.Task):
    try:
        stack = task.get_stack()

        # Format the stack trace into a string
        if stack:
            last_frame = stack[-1]
            formatted_stack_trace = "".join(traceback.format_stack(last_frame))
        else:
            formatted_stack_trace = "No stack trace available"
        return formatted_stack_trace
    except Exception as e:
        logger.exception("Error printing stack")


def background_task(a: Awaitable):
    """Creates a background task that will log any exceptions that occur."""
    task = asyncio.create_task(a)
    task.add_done_callback(background_callback)
    return task


def background_callback(task: asyncio.Task):
    # logger.debug(f"Finished background task {task}")
    try:
        ex = task.exception()
        if ex:
            logger.error(
                f"Exception thrown from background {format_task(task)} task: {ex}"
            )
    except CancelledError:
        pass



import dataclasses
from typing import Optional

from pydantic import BaseModel


class BaseEvent(BaseModel):
    """
    Base class for structured objects used in streams.
    """

    event_name: str
    """The name of the event."""


class SpeechStart(BaseEvent):
    """
    Indicates that speech has been detected in the incoming audio.
    """

    time_since_start: float
    """The time when the speech starts (in seconds) relative to the start of the audio."""

    event_name: str = "speech_start"
    """Always 'speech_start'."""


class SpeechEnd(BaseEvent):
    """
    Indicates that the end of speech has been detected in the incoming audio.
    """

    time_since_start: float
    """The time when the speech ends (in seconds) relative to the start of the audio."""
    event_name: str = "speech_end"
    """Always 'speech_end'"""


class SpeechPartialResult(BaseEvent):
    """
    Provides partial results from a speech recognizer.
    """

    text: str
    """Partial text returned."""
    time_since_start: float
    """The time when the speech ends (in seconds) relative to the start of the audio."""
    event_name: str = "speech_partial_result"
    """Always 'speech_partial_result'"""


class SpeechToTextResult(BaseEvent):
    """
    Provides partial results from a speech recognizer.
    """

    text: str
    """Recognized text"""
    time_since_start: float
    """The time when the speech ends (in seconds) relative to the start of the audio."""
    event_name: str = "speech_to_text_result"
    """Always 'speech_to_text_result'"""


class TimedText(BaseEvent):
    """
    Represents text synchronized to an audio stream. Used for rate-limited output of TextToSpeech.
    """

    text: str
    """The text"""
    duration_in_seconds: float
    """The duration of the audio associated with this text."""
    event_name: str = "timed_text"
    """Always 'timed_text'."""


class AnsweringMachineDetection(BaseEvent):
    """
    Indicates that an answering machine has been detected on a telephone call.
    """

    call_id: str
    """The unique identifier for the call."""
    answered_by: str
    """Who took the call - 'machine' or 'human'."""
    time_since_start: float
    """Time since the start of the call."""
    event_name: str = "amd"
    """Always 'amd'."""

    def is_human(self):
        """
        Check if the response was answered by a human.

        Returns
        -------
        bool
            True if the response was answered by a human, False otherwise.
        """
        return self.answered_by == "human"


class CallStarted(BaseEvent):
    """
    Indicates that a new telephone call has been started.
    """

    call_id: str
    """The unique identifier for the call."""
    stream_id: Optional[str] = None
    """The unique identifier for the audio stream associated with this call."""
    event_name: str = "call_started"
    """Always 'call_started'."""


class CallEnded(BaseEvent):
    """
    Indicates that a telephone call has ended.
    """

    event_name: str = "call_ended"
    """Always 'call_ended'."""



import logging

logger = logging.getLogger(__name__)


def split_webm_buffer(
    audio: bytes, split_before: int, adjust_timestamps: bool = False
) -> bytes:
    """Returns a valid webm stream that strips off roughly `split_before` bytes from the stream.

    Locates the final key frame before `split_before` and creates a new WEBM stream starting at that point.
    Does not do any decoding, so it runs fast.

    Parameters
    ----------
    audio: bytes
        A byte string with the data
    split_before: int
        The split point.  The new webm stream will contain all bytes after the split point, and some from before.
    """

    # Figure out where the header ends
    orig_data_start, _ = find_first_webm_simple_block(audio)

    # Find last key frame before the split.
    new_data_start, _ = find_last_webm_simple_block(
        audio, split_before, keys_frames_only=True, full_frames_only=True
    )

    logger.debug(
        f"Spliting {len(audio)} audio bytes at {new_data_start} (cutoff was {split_before}).  Header is {orig_data_start} bytes"
    )
    return audio[:orig_data_start] + audio[new_data_start:]

    # Original implementation with pyav
    # last_start, size = find_last_webm_simple_block(audio)
    # if len(audio) == last_start + size:
    #     # If the last block was a full block, use the whole thing
    #     logger.debug("Last block was full")
    #     last_start += size
    # usable_audio = audio[:last_start]
    # extra_bytes = audio[last_start:]
    #
    # # Open the input file
    # input_container = av.open(io.BytesIO(usable_audio))
    #
    # # Find the audio stream
    # audio_stream = next((s for s in input_container.streams if s.type == "audio"), None)
    # if audio_stream is None:
    #     raise ValueError("No audio stream found in the WEBM container")
    #
    # # Process and buffer the audio frames
    # packets_since_last_keyframe = []
    # byte_ptr = 0
    # key_frame_start = -1
    # for packet in input_container.demux(audio_stream):
    #     if packet.is_keyframe and byte_ptr < split_before:
    #         packets_since_last_keyframe = []
    #         key_frame_start = byte_ptr
    #     byte_ptr += packet.size
    #     packets_since_last_keyframe.append(packet)
    # logger.debug(
    #     f"Split {len(audio)} audio bytes at {key_frame_start} (cutoff was {split_before}).  {len(packets_since_last_keyframe)} packets included. {len(extra_bytes)} extra bytes"
    # )
    #
    # # Open the output file
    # output_buffer = io.BytesIO()
    # output_container = av.open(output_buffer, mode="w", format="webm")
    # output_container.add_stream(template=audio_stream)
    #
    # # Write the frames from the buffer
    # dts_start = packets_since_last_keyframe[0].dts
    # pts_start = packets_since_last_keyframe[0].pts
    # for packet in packets_since_last_keyframe:
    #     if adjust_timestamps and packet.dts is not None:
    #         packet.dts = packet.dts - dts_start
    #         packet.pts = packet.pts - pts_start
    #     output_container.mux(packet)
    #
    # output_container.close()
    # # PyAV adds a "Cues" section at the end that we don't want.  Cut that off
    # raw = output_buffer.getvalue()
    # pos, size = find_last_webm_simple_block(raw)
    # assert (
    #     len(raw) >= pos + size
    # ), f"Last block should always be a full block.  Internal error."
    # ret = raw[: pos + size] + extra_bytes
    # return ret


def find_first_webm_simple_block(data):
    """
    Searches in a bytes object to find the first occurrence of a "Simple block" webm element.



    `mkvinfo` was really helpful in figuring this out.

    :param data: The bytes object to search.
    :param end: Position to start search backwards from.  If None, starts at the end of the data.
    :return: The position of the '0xA3' byte if found, otherwise -1.
    """
    pos = 0

    while True:
        pos = data.find(b"\xA3", pos)
        if pos == -1:
            # No more occurrences of '0xA3'
            return -1, None

        size = is_webm_simple_block_header(data, pos)
        if size != -1:
            return pos, size

        # Update the search range to exclude the current '0xA3' position
        pos += 1


def find_last_webm_simple_block(
    data,
    end: int = None,
    keys_frames_only: bool = False,
    full_frames_only: bool = False,
):
    """
    Searches backwards in a bytes object to find the last occurrence of
    a "Simple block" webm element.

    Looks for "a3", followed by a 1 or 2 byte length, A variable length in less than 5, then skip two bytes and look for 80 or 0.

    `mkvinfo` was really helpful in figuring this out.

    :param data: The bytes object to search.
    :param end: Position to start search backwards from.  If None, starts at the end of the data.
    :return: A tuple of the position of the '0xA3' byte if found, otherwise -1 and then the size (None is pos is -1)
    """
    l = end or len(data)
    pos = l

    while True:
        pos = data.rfind(b"\xA3", 0, pos)
        if pos == -1:
            # No more occurrences of '0xA3'
            return -1, None
        # logger.debug(f"Checking {pos:x}")
        size = is_webm_simple_block_header(data, pos, l, is_key_frame=keys_frames_only)
        if size != -1:
            # logger.debug(f"Found block at {pos}")
            if not full_frames_only or pos + size <= l:
                return pos, size

        # Update the search range to exclude the current '0xA3' position
        pos -= 1


def is_webm_simple_block_header(
    data, pos, end: int = None, is_key_frame: bool = False
) -> int:
    """Returns the length of the header or -1 if it isn't a header.

    Looks for "a3", followed by a 1 or 2 byte length, A variable length in less than 5, then skip two bytes and look for 80 or 0.
    """
    l = end or len(data)
    if data[pos] != 0xA3:
        logger.debug("Initial byte didn't match")
        return -1
    if pos + 5 < l:
        size, size_bytes = read_ebml_size(data, pos + 1)
        if size is not None:
            track, track_bytes = read_ebml_size(data, pos + 1 + size_bytes)
            if (
                track is not None and track < 10
            ):  # Something is fishy if there are more than 10 tracks
                flag_pos = pos + 1 + size_bytes + track_bytes + 2
                if flag_pos < l:
                    flag = data[pos + 1 + size_bytes + track_bytes + 2]
                    key_frame = flag | 0x80
                    if (
                        flag - key_frame == 0x00
                    ):  # We expect the only flag set is key_frame.
                        if key_frame or not is_key_frame:
                            return size + 1 + size_bytes
    return -1


def read_ebml_size(data, position, max_bytes=2):
    """
    Reads a variable-length integer from an EBML stream.

    Returns None as the value if you hit the end of stream or the value is larger than max_bytes bytes in length.

    :param data: The bytes object containing the EBML data.
    :param position: The position in the data where the size field starts.
    :param max_bytes: The maximum number of expected bytes.  Throws ValueError if this is off.
    :return: The interpreted size value and the number of bytes.
    """
    # Extract the first byte
    first_byte = data[position]

    # Count leading zeros to determine the number of bytes in the size field
    num_bytes = 1
    for bit in range(7, -1, -1):
        if (first_byte >> bit) & 1:
            break
        num_bytes += 1

    if num_bytes > max_bytes:
        return None, num_bytes

    # If num_bytes is 0, it means the size is in this single byte
    if num_bytes == 1:
        return first_byte & 0x7F, 1  # Mask out the marker bit

    # Read the size field bytes
    if len(data) < position + num_bytes:
        return None, num_bytes
    size_bytes = data[position : position + num_bytes]

    # Calculate the size value
    size_value = 0
    for i, byte in enumerate(size_bytes):
        if i == 0:
            # For the first byte, mask out the marker bits
            size_value = byte & (0xFF >> (num_bytes + 1))
        else:
            size_value = (size_value << 8) | byte

    return size_value, num_bytes



import asyncio
import inspect
import logging
import time
from typing import AsyncIterator

import asyncstdlib

from voice_stream import (
    chunk_bytes_step,
    async_init_step,
    extract_value_step,
    substream_step,
)
from voice_stream.audio.async_wav_file import (
    AsyncMuLawStreamWriter,
    AsyncMuLawStreamReader,
)
from voice_stream.audio.audio_mp3 import find_frame_boundaries, calculate_split_points
from voice_stream.audio.audio_ogg import OggPage, OpusIdPacket
from voice_stream.audio.audio_utils import (
    AudioFormatError,
    AudioFormat,
    get_audio_length,
)
from voice_stream.types import resolve_awaitable_or_obj, AwaitableOrObj, map_future

logger = logging.getLogger(__name__)


async def wav_mulaw_file_source(
    filename: str, chunk_size: int = 4096
) -> AsyncIterator[bytes]:
    """
    Data flow source that reads audio bytes from a wav file.

    Parameters
    ----------
    filename : str
        Name of the audio file
    chunk_size : int
        Number of bytes to read at one time from the file.  Passing 0 indicates the whole file should be read at once.

    Returns
    -------
    AsyncIterator[bytes]
        A stream of audio bytes.


    Notes
    -------
    - The WAV header will be removed, so the data being passed will only include audio samples.
    """
    f = AsyncMuLawStreamReader(filename, chunk_size=chunk_size)
    try:
        await f.open()
        while True:
            data = await f.read()
            if not data:
                break
            # logger.debug(f"Read {len(data)} bytes from {filename}")
            yield data
    finally:
        await f.close()


async def wav_mulaw_file_sink(
    async_iter: AsyncIterator[bytes], filename: AwaitableOrObj[str]
) -> None:
    """
    Data flow sink that writes telephone audio (8Khz mu-law encoded audio) to a WAV file.

    Parameters
    ----------
    async_iter : str
        A stream containing the audio data to write.
    filename : str
        Name of the audio file.  Can be an `Awaitable[str]` if the filename isn't known at creation time (for example if it is generated based on data in the stream).

    Notes
    -------
    - Assumes only audio data is passed.  A WAV header will be placed before the audio to properly format the file.
    """
    f = None
    try:
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for message in owned_aiter:
                # We wait to resolve the filename until we have a message to write
                if f is None:
                    filename = await resolve_awaitable_or_obj(filename)
                    f = AsyncMuLawStreamWriter(filename)
                    await f.open()
                if message is None:
                    break
                # logger.debug(f"Wrote {len(message)} bytes to {filename}")
                await f.write(message)
    finally:
        if f is not None:
            await f.close()
        logger.info(f"Closed {filename}")


async def mp3_chunk_step(
    async_iter: AsyncIterator[bytes], chunk_size: AwaitableOrObj[int]
) -> AsyncIterator[bytes]:
    """
    Data flow step that splits incoming MP3 data on MP3 frame boundaries.

    Takes incoming audio data and splits it into smaller chunks based on MP3 frame boundaries.  Attempts to keep
    outgoing bytes smaller than `chunk_size` but may be larger if an individual frame can't fit into a chunk.

    Parameters
    ----------
    async_iter :
        The MP3 data to be split into chunks

    chunk_size : int
        The maximum allowable chunk size.

    Returns
    -------
    AsyncIterator[bytes]
        The incoming data split into smaller chunks which are all complete MP3 frames.

    Notes
    -----
    - If an individual frame is larger than the chunk_size, it will not be split and the chunk can exceed chunk_size.
    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for data in owned_aiter:
            resolved_chunk_size = await resolve_awaitable_or_obj(chunk_size)
            boundaries = find_frame_boundaries(data)
            splits = calculate_split_points(boundaries, len(data), resolved_chunk_size)
            split_start = 0
            for split_end in splits:
                yield data[split_start:split_end]
                split_start = split_end
            assert split_end == len(data)


async def ogg_page_separator_step(
    async_iter: AsyncIterator[bytes],
) -> AsyncIterator[bytes]:
    """
    Data flow step that splits incoming OGG data into distinct pages.

    Takes in a stream of bytes from an OGG media file and outputs bytes ensuring that each output is a complete page.
    Checks if the last page in each chunk sent is a full page, if so, it sends it.

    Parameters
    ----------
    async_iter
        Bytes from an OGG media file.

    Returns
    -------
    AsyncIterator[bytes]
        Bytes objects, each representing a full OGG page.

    Notes
    -----
    - If the data passed in contains a partial page at the end, that page data will be buffered until the next input.
    """
    buffer = b""
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for data in owned_aiter:
            concatenated = buffer + data
            if concatenated[:4] != b"OggS":
                raise AudioFormatError(
                    f"Bytes didn't start with a proper OggS head.  Expected 'OggS, got {concatenated[:4]}"
                )
            position = concatenated.rfind(b"OggS")
            if OggPage.is_full_page(concatenated, position):
                position = len(concatenated)
            buffer = concatenated[position:]
            if position > 0:
                yield concatenated[:position]
    if len(buffer) > 0:
        yield buffer


async def ogg_concatenator_step(
    async_iter: AsyncIterator[bytes],
) -> AsyncIterator[bytes]:
    """
    Data flow step that concatenates multiple OGG streams into one.

    With files in OGG format, you can't concatenate two different streams simply by concatenating the bytes.
    This step performs the necessary operations to concatenate streams.  It assumes data is coming in chunked into
    full OGG pages (the way :func:`~voice_stream.audio.ogg_page_separator_step` outputs it).

    Parameters
    ----------
    async_iter
        OGG pages as bytes objects

    Returns
    -------
    AsyncIterator[bytes]
        OGG pages, updated so that they form a single consistent stream.

    Notes
    -----
    - This could be done more efficiently by using numpy to modify the buffer in place.
    """
    sequence_offset = 0
    granule_offset = 0
    header = None

    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for data in owned_aiter:
            # Iterate through each ogg page and update.
            position = 0
            pages = []
            while position < len(data):
                page = OggPage.from_data(data, position)
                page_header_type = page.header_type
                position += page.page_length
                if page.page_sequence_number == 0:
                    if not OpusIdPacket.is_opus_encoded(
                        data, page.offset + page.header_length
                    ):
                        raise AudioFormatError(
                            "OGG file is not OPUS encoded.  Only Opus is currently supported."
                        )
                    new_header = OpusIdPacket.from_data(
                        data, page.offset + page.header_length
                    )
                    if header is not None:
                        if (
                            header.output_channel_count
                            != new_header.output_channel_count
                            or header.input_sample_rate != new_header.input_sample_rate
                            or header.output_gain != new_header.output_gain
                            or header.channel_mapping_family
                            != new_header.channel_mapping_family
                        ):
                            raise AudioFormatError(
                                "Cant concatenate OGG streams with different headers."
                            )
                if sequence_offset and page.page_sequence_number < 2:
                    logger.debug("skipping opus header for later sequence")
                    continue
                if sequence_offset:
                    page = page.update(
                        new_header_type=0,
                        granule_offset=granule_offset,
                        sequence_offset=sequence_offset,
                    )
                    # page.log()
                if page_header_type == 4:
                    sequence_offset = sequence_offset + page.page_sequence_number
                    granule_offset = granule_offset + page.granule
                pages.append(page)
            logger.debug(f"Adjusted {len(pages)} OGG pages")
            adjusted = b"".join(_.get_bytes() for _ in pages)
            yield adjusted


def audio_rate_limit_step(
    async_iter: AsyncIterator[bytes],
    audio_format: AwaitableOrObj[AudioFormat],
    buffer_seconds: float,
):
    """
    Data flow step that rate-limits the audio data coming in.

    This step takes in audio data and produces the same audio data with delays introduced so that the downstream
    iterator only gets `buffer_seconds` worth of audio at once.  Rate-limiting provides the ability to stop the audio
    stream due to an interruption or other event.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator returning bytes of audio data.

    audio_format : AwaitableOrObj[voice_stream.audio.AudioFormat]
       The format of the audio data.  Can be an Awaitable if the format isn't known when the step is created.

    buffer_seconds : float
        The amount of audio to pass to the downstream iterator.

    Returns
    -------
    audio
        The same audio bytes that came in, but rate-limited so that the downstream consumer only gets `buffer_seconds` worth of audio.

    Raises
    ------
    AudioFormatError
        If the audio format is not supported.

    Notes
    ------
    - This function will break up long chunks of data in a format-specific way to perform the rate-limiting.
    """

    def init(async_iter, af):
        if af == AudioFormat.WAV_MULAW_8KHZ:
            SAMPLE_RATE = 8000
            audio = chunk_bytes_step(async_iter, int(buffer_seconds * SAMPLE_RATE))
            audio = raw_audio_rate_limit_step(
                audio, SAMPLE_RATE, buffer_seconds=buffer_seconds
            )
        elif af == AudioFormat.OGG_OPUS:
            audio = _opus_rate_limit_step(async_iter, buffer_seconds=buffer_seconds)
        elif af == AudioFormat.MP3:
            audio = _mp3_rate_limit_step(async_iter, buffer_seconds=buffer_seconds)
        else:
            raise AudioFormatError(f"Unsupported audio format: {audio_format}")
        return audio

    async def async_init(async_iter):
        af = await resolve_awaitable_or_obj(audio_format)
        return init(async_iter, af)

    if inspect.isawaitable(audio_format):
        return async_init_step(async_iter, async_init)
    else:
        return init(async_iter, audio_format)


def _opus_rate_limit_step(async_iter: AsyncIterator[bytes], buffer_seconds: float):
    def get_bytes_per_second(data):
        duration = get_audio_length(AudioFormat.OGG_OPUS, data)
        length_in_bytes = len(data)
        return int(length_in_bytes / duration)

    def opus_substream(async_iter: AsyncIterator[bytes]):
        stream, bytes_per_second_f = extract_value_step(
            async_iter, get_bytes_per_second
        )
        chunk_size_f = map_future(
            bytes_per_second_f,
            lambda bytes_per_second: int(buffer_seconds * bytes_per_second / 2),
        )
        stream = chunk_bytes_step(stream, chunk_size_f)
        # stream = log_step(stream, "Audio Chunk", lambda x: len(x))
        stream = raw_audio_rate_limit_step(
            stream, bytes_per_second_f, buffer_seconds=buffer_seconds
        )
        stream = ogg_page_separator_step(stream)
        # stream = log_step(stream, "Full OGG Page", lambda x: len(x))
        return stream

    out = substream_step(async_iter, opus_substream)
    return out


def _mp3_rate_limit_step(async_iter: AsyncIterator[bytes], buffer_seconds: float):
    def get_bytes_per_second(data):
        seconds = get_audio_length(AudioFormat.MP3, data)
        length_in_bytes = len(data)
        return int(length_in_bytes / seconds)

    def mp3_substream(async_iter: AsyncIterator[bytes]):
        stream, bytes_per_second_f = extract_value_step(
            async_iter, get_bytes_per_second
        )
        chunk_size_f = map_future(
            bytes_per_second_f,
            lambda bytes_per_second: int(buffer_seconds * bytes_per_second / 2),
        )
        stream = mp3_chunk_step(stream, chunk_size_f)
        stream = raw_audio_rate_limit_step(
            stream, bytes_per_second_f, buffer_seconds=buffer_seconds
        )
        return stream

    out = substream_step(async_iter, mp3_substream)
    return out


async def raw_audio_rate_limit_step(
    async_iter: AsyncIterator[bytes],
    bytes_per_second: AwaitableOrObj[int],
    buffer_seconds: AwaitableOrObj[float],
) -> AsyncIterator[bytes]:
    """
    Data flow step that performs rate-liming on chunks of audio data coming in.

    This step rate limits input objects based on a given sample rate.  Generally, using :func:`~voice_stream.audio.audio_rate_limit_step`
     is preferred to using this step, as that step handles the details of different audio formats.  This step does not
     break up long chunks or handle differing formats.  It takes in bytes objects, determines the length of the audio based
     on `bytes_per_second` and outputs the identical chunks it got with a delay.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        The input audio data
    bytes_per_second: AwaitableOrObj[int],
        The playback rate of the audio in bytes.  This is used to compute the duration of the audio based on the length of the byte array.
    buffer_seconds : float
        The number of seconds of audio left before we send the next chunk.

    Note
    ----
    Usually, you will want to put a max size chunk step in before this.
    """
    last_send = time.perf_counter()
    queued_audio_seconds = 0

    def compute_remaining(now):
        """Compute the amount of audio remaining to be played in seconds."""
        return max(0, queued_audio_seconds - (now - last_send))

    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            resolved_buffer_seconds = await resolve_awaitable_or_obj(buffer_seconds)
            now = time.perf_counter()
            # Compute the amount of audio remaining to be played in seconds.
            remaining_audio_seconds = compute_remaining(now)
            # If we have more than the buffer, sleep until we hit the buffer limit
            if remaining_audio_seconds > resolved_buffer_seconds:
                await asyncio.sleep(remaining_audio_seconds - resolved_buffer_seconds)
                # We don't know how long we actually slept.
                now = time.perf_counter()
                remaining_audio_seconds = compute_remaining(now)
            resolved_bytes_per_second = await resolve_awaitable_or_obj(bytes_per_second)
            queued_audio_seconds = (
                remaining_audio_seconds + len(item) / resolved_bytes_per_second
            )
            last_send = now
            yield item



"""
`FastAPI <https://fastapi.tiangolo.com/>`_ is a high performance web framework.  VoiceStream provides integrations to
send and receive audio using websockets using FastAPI.
"""

import logging
from typing import AsyncIterator, Union

import asyncstdlib

try:
    from fastapi import WebSocket
except ImportError:
    raise ImportError(
        "Could not import fastapi python package. "
        "Please install it with `pip install fastapi`."
    )

from starlette.websockets import WebSocketDisconnect, WebSocketState

logger = logging.getLogger(__name__)


async def fastapi_websocket_text_source(websocket: WebSocket) -> AsyncIterator[str]:
    """
    Data flow source for receiving text messages from a FastAPI WebSocket connection.

    This function facilitates the handling of incoming text messages from a client connected through
    a FastAPI WebSocket. It continuously listens for messages and yields them as they arrive. The
    function handles the WebSocket connection lifecycle by accepting the connection and managing
    disconnection events.

    Parameters
    ----------
    websocket : WebSocket
        An instance of WebSocket from FastAPI. This is the WebSocket connection through which the
        function will receive text messages.

    Returns
    -------
    AsyncIterator[str]
        An asynchronous iterator that yields incoming text messages as strings.

    Notes
    -----
    - The function uses an infinite loop to listen for messages. It exits the loop and stops iterating
      when a WebSocketDisconnect exception occurs.
    - This source will call `accept` on the websocket if it has not already been called.


    Examples
    --------
        >>> from fastapi import FastAPI, WebSocket
        >>> import asyncio
        >>>
        >>> app = FastAPI()
        >>>
        >>> @app.websocket("/ws")
        >>> async def websocket_endpoint(websocket: WebSocket):
        >>>     stream = fastapi_websocket_text_source(websocket)
        >>>     stream = map_step(stream, lambda x: "Echo: "+x)
        >>>     await fastapi_websocket_text_sink(stream, websocket)
    """
    if websocket.client_state == WebSocketState.CONNECTING:
        await websocket.accept()
    try:
        while True:
            yield await websocket.receive_text()
    except WebSocketDisconnect:
        pass


async def fastapi_websocket_bytes_source(websocket: WebSocket) -> AsyncIterator[bytes]:
    """
    Data flow source for binary data received from a FastAPI WebSocket.

    This function continuously listens to a WebSocket connection, yielding binary data as
    it's received. The iterator will handle the connection establishment and will
    terminate gracefully if the WebSocket disconnects.

    Parameters
    ----------
    websocket : WebSocket
        The FastAPI WebSocket connection from which to receive binary data.

    Returns
    -------
    bytes
        Binary data received from the WebSocket.

    Examples
    --------
    >>> from fastapi import FastAPI, WebSocket
    >>> import asyncio
    >>>
    >>> app = FastAPI()
    >>>
    >>> @app.websocket("/ws")
    >>> async def websocket_endpoint(websocket: WebSocket):
    >>>     stream = fastapi_websocket_text_source(websocket)
    >>>     stream = map_step(stream, lambda x: "Echo: "+x)
    >>>     await fastapi_websocket_text_sink(stream, websocket)

    Notes
    -----
    This function is designed to be used with FastAPI's WebSocket support.
    It requires an established WebSocket connection passed as the parameter.

    This source will call `accept` on the websocket if it has not already been called.

    See Also
    --------
    WebSocket.receive_bytes : Method of FastAPI's WebSocket to receive binary data.

    """
    if websocket.client_state == WebSocketState.CONNECTING:
        await websocket.accept()
    try:
        while True:
            yield await websocket.receive_bytes()
    except WebSocketDisconnect:
        pass


async def fastapi_websocket_json_source(websocket: WebSocket) -> AsyncIterator[dict]:
    """
    Data flow source for JSON data received from a FastAPI WebSocket.

    This function continuously listens to a WebSocket connection, yielding JSON data (in the
    form of Python dictionaries) as it's received. The iterator manages the connection
    establishment and will terminate gracefully if the WebSocket disconnects.

    Parameters
    ----------
    websocket : WebSocket
        The WebSocket connection from which to receive JSON data.

    Returns
    -------
    AsyncIterator[dict]
        JSON data received from the WebSocket, parsed into a Python dictionary.

    Examples
    --------
    >>> from fastapi import FastAPI, WebSocket
    >>> import asyncio
    >>>
    >>> app = FastAPI()
    >>>
    >>> @app.websocket("/ws")
    >>> async def websocket_endpoint(websocket: WebSocket):
    >>>     stream = fastapi_websocket_json_source(websocket)
    >>>     stream = log_step(stream, "JSON received")
    >>>     await empty_step(stream, websocket)

    Notes
    -----
    This function is specifically tailored for use with FastAPI's WebSocket support.
    It requires an active WebSocket connection as the parameter.

    This source will call `accept` on the websocket if it has not already been called.

    See Also
    --------
    WebSocket.receive_json : Method of FastAPI's WebSocket to receive JSON data.

    """
    if websocket.client_state == WebSocketState.CONNECTING:
        await websocket.accept()
    try:
        while True:
            yield await websocket.receive_json()
    except WebSocketDisconnect:
        pass


async def fastapi_websocket_text_sink(
    async_iter: AsyncIterator[Union[str, dict]], websocket: WebSocket
) -> None:
    """
    Data flow sink to send data to a FastAPI WebSocket connection.

    This function takes an asynchronous iterator, which can yield either strings, Pydantic objects, or
    dictionaries, and sends each item to a specified WebSocket. If the item is a dictionary,
    it is sent as a JSON message. Otherwise, it is sent as a text message.

    Parameters
    ----------
    async_iter : AsyncIterator[Union[str, dict]]
        An asynchronous iterator that yields either strings or dictionaries.
    websocket : WebSocket
        The FastAPI WebSocket connection to which the data will be sent.

    Examples
    --------
    >>> from fastapi import FastAPI, WebSocket
    >>> import asyncio
    >>>
    >>> app = FastAPI()
    >>>
    >>> @app.websocket("/ws")
    >>> async def websocket_endpoint(websocket: WebSocket):
    >>>     stream = fastapi_websocket_json_source(websocket)
    >>>     stream = log_step(stream, "JSON received")
    >>>     await fastapi_websocket_text_sink(stream, websocket)

    Notes
    -----
    - This function is useful for streaming data from an asynchronous source to a WebSocket
      client. It supports both text and JSON formats, making it versatile for various types
      of data communication in a FastAPI application.
    - This sink will call `accept` on the websocket if it has not already been called.

    See Also
    --------
    WebSocket.send_json : Method of FastAPI's WebSocket to send JSON data.
    WebSocket.send_text : Method of FastAPI's WebSocket to send text data.

    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        if websocket.client_state == WebSocketState.CONNECTING:
            await websocket.accept()
        async for message in owned_aiter:
            if isinstance(message, dict):
                await websocket.send_json(message)
            elif hasattr(message, "model_dump"):
                json = message.model_dump()
                logger.info(f"Sending {json}")
                await websocket.send_json(json)
            else:
                await websocket.send_text(message)


async def fastapi_websocket_bytes_sink(
    async_iter: AsyncIterator[bytes], websocket: WebSocket
) -> None:
    """
    Data flow sink to send bytes to a FastAPI websocket.

    This function takes an asynchronous iterator that yields bytes, and for each item in this iterator,
    it sends the item to the specified FastAPI WebSocket connection.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator that yields bytes. Each byte sequence received from this iterator
        will be sent to the websocket.
    websocket : WebSocket
        The FastAPI WebSocket instance to which the byte sequences will be sent.

    Returns
    -------
    None
        This function does not return any value.

    Examples
    --------
    >>> from fastapi import FastAPI, WebSocket
    >>> import asyncio
    >>>
    >>> app = FastAPI()
    >>>
    >>> @app.websocket("/ws")
    >>> async def websocket_endpoint(websocket: WebSocket):
    >>>     stream = fastapi_websocket_bytes_source(websocket)
    >>>     stream = log_step(stream, "# Bytes received", lambda x: len(x))
    >>>     await fastapi_websocket_bytes_sink(stream, websocket)

    Notes
    -----
    - It's essential that the `async_iter` provided is an asynchronous iterator yielding byte sequences.
    - The function will continue sending data until the iterator is exhausted.
    - In case of any interruption or when the iterator is exhausted, it ensures proper cleanup.
    - This sink will call `accept` on the websocket if it has not already been called.
    """
    try:
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            if websocket.client_state == WebSocketState.CONNECTING:
                await websocket.accept()
            async for message in owned_aiter:
                await websocket.send_bytes(message)
    finally:
        logger.debug("fastapi_websocket_bytes_sink done.")



# Google Speech by itself limits streams to 5 minutes, so you have to play games to have endless recognition.
# See https://github.com/GoogleCloudPlatform/python-docs-samples/blob/main/speech/microphone/transcribe_streaming_infinite.py
import logging
import sys
import time
from asyncio import Event
from typing import AsyncIterator, Callable, Union, Optional

import asyncstdlib
from google.api_core.exceptions import Cancelled
from google.cloud.speech_v1 import (
    StreamingRecognizeRequest as StreamingRecognizeRequestV1,
)
from google.cloud.speech_v1 import (
    StreamingRecognizeResponse as StreamingRecognizeResponseV1,
)
from google.cloud.speech_v2 import StreamingRecognizeResponse, StreamingRecognizeRequest

from voice_stream import (
    byte_buffer_step,
    chunk_bytes_step,
    queue_sink,
    queue_source,
    log_step,
    concat_step,
    async_init_step,
    QueueWithException,
    EndOfStreamMarker,
)
from voice_stream.audio import AudioFormat
from voice_stream.audio.audio_webm import split_webm_buffer
from voice_stream.core import (
    single_source,
    _chunk_bytes,
    fork_step,
    text_file_sink,
    map_step,
    binary_file_sink,
)
from voice_stream.events import BaseEvent, SpeechToTextResult
from voice_stream.types import background_task, format_current_task

logger = logging.getLogger(__name__)


def _wrap_streaming_recognize(stream, speech_async_client):
    """streaming_recognize doesn't exit cleanly or have good error handling.

    This fixes it so that when the input stream stops cleanly, the output stream stops cleanly, and when the
    input stream has an exception, it is propagated to the output stream.
    """
    exc_info = None

    async def input_wrap(async_iter):
        try:
            async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
                async for item in owned_aiter:
                    yield item
        except Exception as e:
            nonlocal exc_info
            exc_info = sys.exc_info()

    async def output_wrap(async_iter):
        try:
            async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
                async for item in owned_aiter:
                    yield item
        except Cancelled as e:
            # We always get a cancelled out of the grpc, even for a clean stop in the audio.
            # If there was an input error, propagate that.  Otherwise, exit cleanly.
            pass
        if exc_info:
            exc_type, exc_value, exc_traceback = exc_info
            raise exc_value

    stream = input_wrap(stream)
    stream = async_init_step(stream, speech_async_client.streaming_recognize)
    stream = output_wrap(stream)
    return stream


class EndlessStream:
    MAX_STREAM_SIZE = 25600

    class StreamTimeoutException(Exception):
        pass

    def __init__(
        self,
        speech_async_client,
        initial_config: Union[StreamingRecognizeRequest, StreamingRecognizeRequestV1],
        map_audio: Callable[
            [bytes], Union[StreamingRecognizeRequest, StreamingRecognizeRequestV1]
        ],
        map_event: Callable[
            [Union[StreamingRecognizeResponse, StreamingRecognizeResponseV1], float],
            Optional[BaseEvent],
        ],
        audio_format: AudioFormat,
        stream_reset_timeout_secs: int = 240,
        max_resets: int = 4,
    ):
        self.speech_async_client = speech_async_client
        self.initial_config = initial_config
        self.map_audio = map_audio
        self.map_event = map_event
        self.audio_format = audio_format
        self.stream_reset_timeout_secs = stream_reset_timeout_secs
        self.max_resets = max_resets

        # Modifiable state
        self.new_stream = True
        self.stream_start_time = None
        self.duration_of_previous_streams = 0
        self.last_final_result_end_time = 0
        self.restart_counter = 0
        self.output_queue = QueueWithException()

    def step(self, async_iter):
        background_task(self._run_streams(async_iter))
        return queue_source(self.output_queue)

    async def _run_streams(self, async_iter: AsyncIterator[bytes]):
        queue_audio_task = None
        try:
            # Enqueue the input audio
            stream = async_iter
            stream = byte_buffer_step(stream)
            stream = chunk_bytes_step(stream, EndlessStream.MAX_STREAM_SIZE)
            stream = self.input_iterator(stream)
            input_queue = QueueWithException()
            queue_audio_task = background_task(queue_sink(stream, input_queue))

            while self.restart_counter < self.max_resets:
                cancel_event = Event()
                try:
                    stream = queue_source(input_queue, cancel_event)
                    # stream = log_step(stream, "Audio", lambda x: f"{format_current_task()}")
                    # stream, audio_out = fork_step(stream)
                    # audio_out = map_step(audio_out, lambda x:x.audio_content if hasattr(x, "audio_content") else x.audio)
                    # background_task(binary_file_sink(audio_out, f"{self.restart_counter}.webm"))
                    # background_task(text_file_sink(audio_out, f"{self.restart_counter}.ndjson"))
                    config = single_source(self.initial_config)
                    config = log_step(
                        config,
                        f"Recognize started #{self.restart_counter+1}",
                        lambda x: "",
                    )
                    stream = concat_step(config, stream)
                    # stream = log_step(
                    #     stream, f"Audio to SR #{self.restart_counter+1}", lambda x: f"", every_nth_message=25
                    # )
                    stream = _wrap_streaming_recognize(stream, self.speech_async_client)
                    stream = self.output_iterator(stream)
                    logger.debug("Initiating stream")
                    await queue_sink(
                        stream,
                        self.output_queue,
                        send_end_of_stream=False,
                        propagate_exceptions=False,
                    )
                    logger.debug("Endless stream completed successfully")
                    return
                except EndlessStream.StreamTimeoutException:
                    logger.debug("Endless stream timeout exception, restarting stream")
                    self.restart_counter += 1
                    cancel_event.set()
                    # Drain any queued audio from the input queue (it will be replayed with the new stream).
                    logger.debug(
                        f"Clearing {input_queue.qsize()} remaining items from the queue"
                    )
                    while not input_queue.empty():
                        input_queue.get_nowait()
                    if queue_audio_task.done():
                        return
                    self.new_stream = True
            logger.debug(
                f"Exceeded maximum reset count of {self.max_resets}.  Stopping stream."
            )
        except Exception as e:
            logger.exception(f"Exception during Google Recognize.")
            await self.output_queue.set_exception(e)
        finally:
            if queue_audio_task:
                queue_audio_task.cancel()
            if not self.output_queue.exception:
                await self.output_queue.put(EndOfStreamMarker)

    async def input_iterator(self, async_iter):
        duration_of_replay_buffer = 0
        replay_buffer = []
        self.stream_start_time = time.perf_counter()
        count = 0
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for item in owned_aiter:
                if self.new_stream and replay_buffer:
                    # If we are starting a new stream, replay from our last final result in the old stream.
                    seconds_per_chunk = (
                        duration_of_replay_buffer + self.stream_reset_timeout_secs
                    ) / len(replay_buffer)
                    logger.debug(
                        f"Restarting last_end: {self.last_final_result_end_time} duration_of_previous_streams {self.duration_of_previous_streams} duration_of_replay_buffer {duration_of_replay_buffer} seconds_per_chunk: {seconds_per_chunk}"
                    )

                    n_chunks_fully_processed = round(
                        (
                            self.last_final_result_end_time
                            - self.duration_of_previous_streams
                            - duration_of_replay_buffer
                        )
                        / seconds_per_chunk
                    )

                    self.duration_of_previous_streams += (
                        n_chunks_fully_processed * seconds_per_chunk
                    )

                    duration_of_replay_buffer = (
                        len(replay_buffer) - n_chunks_fully_processed
                    ) * seconds_per_chunk
                    logger.debug(
                        f"Restarting stream with {duration_of_replay_buffer} seconds of replay.  Previous streams were {self.duration_of_previous_streams} seconds."
                    )

                    replay_audio = self.split_stream(
                        replay_buffer, n_chunks_fully_processed
                    )
                    replay_buffer = []
                    count += 1
                    for c in _chunk_bytes(replay_audio, EndlessStream.MAX_STREAM_SIZE):
                        replay_buffer.append(c)
                        ret = self.map_audio(c)
                        yield ret

                if self.new_stream:
                    logger.debug(
                        f"New stream initiating with {len(item)} bytes of audio"
                    )
                self.new_stream = False
                replay_buffer.append(item)
                yield self.map_audio(item)
        logger.debug("Reached the end of the input audio.")

    async def output_iterator(self, async_iter: AsyncIterator):
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for item in owned_aiter:
                event = self.map_event(item, self.duration_of_previous_streams)
                if isinstance(event, SpeechToTextResult):
                    self.last_final_result_end_time = event.time_since_start

                yield event

                now = time.perf_counter()
                if now - self.stream_start_time > self.stream_reset_timeout_secs:
                    self.stream_start_time = now
                    raise EndlessStream.StreamTimeoutException()

    def split_stream(self, frame_list: list[bytes], split_point: int):
        if (
            self.audio_format == AudioFormat.WAV_MULAW_8KHZ
            or self.audio_format == AudioFormat.MP3
        ):
            return b"".join(frame_list[split_point:])
        elif self.audio_format == AudioFormat.WEBM_OPUS:
            byte_cutoff = 0
            for i in range(split_point):
                byte_cutoff += len(frame_list[i])
            replay_audio = b"".join(frame_list)
            return split_webm_buffer(replay_audio, byte_cutoff)
        elif self.audio_format == AudioFormat.OGG_OPUS:
            raise ValueError(
                f"Splitting OGG requires updating all future packets.  Not really feasible."
            )
        else:
            raise ValueError(
                f"Audio format not supported for split {self.audio_format}"
            )



"""
`Twilio <https://www.twilio.com>`_ provides an API for making and receiving telephone calls.  You can use Twilio along
with VoiceStream to automate calls with an LLM.
"""

from __future__ import annotations

import asyncio
import base64
import dataclasses
import logging
from typing import AsyncIterator, Tuple, Optional, Dict, Any, Callable, Union, Coroutine

import asyncstdlib

from voice_stream.core import (
    AwaitableOrObj,
    partition_step,
    map_step,
    map_str_to_json_step,
    filter_step,
    fork_step,
    extract_value_step,
    merge_step,
    queue_source,
)
from voice_stream.events import BaseEvent, CallStarted, CallEnded
from voice_stream.types import map_future, resolve_awaitable_or_obj

logger = logging.getLogger(__name__)


def twilio_split_media_step(
    async_iter: AsyncIterator[dict],
) -> Tuple[AsyncIterator[dict], AsyncIterator[dict]]:
    """
    Data flow step that splits a Twilio stream into two separate streams, one for media
    stream and another for control messages.

    Parameters
    ----------
    async_iter : AsyncIterator[dict]
        An asynchronous iterator over Twilio stream events. Each event is represented
        as a dictionary.

    Returns
    -------
    Tuple[AsyncIterator[dict], AsyncIterator[dict]]
        A tuple of two asynchronous iterators. The first iterator yields dictionaries
        of media stream events, and the second yields dictionaries of control messages.

    Notes
    -----
    - The function relies on the presence of an "event" key in each dictionary to
      determine the type of the message.
    - This function is useful for handling different types of events in a Twilio stream
      separately, especially in applications that need to process media and control
      messages differently.

    Examples
    --------
    >>> media_stream, control_messages = twilio_split_media_step(async_iter)
    """
    return partition_step(async_iter, lambda x: x["event"] == "media")


class TwilioSequenceError(ValueError):
    """Indicates a problem in the sequence numbers for messages received from Twilio."""

    pass


async def twilio_check_sequence_step(
    async_iter: AsyncIterator[dict],
) -> AsyncIterator[dict]:
    """
    Data flow step that verifies that no Twilio websocket messages were lost.

    This step checks each message in the provided asynchronous iterator for a
    sequence number and validates that the messages are in the correct sequence.
    If a message is found to be out of sequence or missing a sequence number, a
    TwilioSequenceError is raised.

    Parameters
    ----------
    async_iter : AsyncIterator[dict]
        An asynchronous iterator over Twilio websocket messages, where each message
        is expected to be a dictionary containing a 'sequenceNumber' key.

    Yields
    ------
    AsyncIterator[dict]
        An asynchronous iterator yielding the same Twilio websocket messages after
        verifying their sequence.

    Raises
    ------
    TwilioSequenceError
        If a message is missing a sequence number or if the sequence of messages is
        found to be out of order.

    Notes
    -----
    - The function starts by expecting a sequence number of 1 and increments this
      expectation with each message.
    """
    expected = 1
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if "sequenceNumber" not in item:
                raise TwilioSequenceError(f"Message {item} has no sequence number")
            if int(item["sequenceNumber"]) != expected:
                raise TwilioSequenceError(
                    f"Expected sequence number {expected}, got {item['sequenceNumber']}"
                )
            expected += 1
            yield item


async def twilio_close_on_stop_step(
    async_iter: AsyncIterator[dict],
    close_func: Callable[[], Union[None, Coroutine[Any, Any, None]]],
) -> AsyncIterator[dict]:
    """
    Data flow step that calls a close function when a stop message is received.

    Receives Twilio websocket messages and checks for a 'stop' event.  Upon receiving a 'stop' message, it triggers
    a closure function (typically to close the websocket) and continues to yield the remaining messages.

    Parameters
    ----------
    async_iter : AsyncIterator[dict]
        An asynchronous iterator over Twilio websocket messages.
    close_func : Callable[[], None]
        A asynchronous callable function that is executed when a 'stop' event is received. This function is awaited
        when the 'stop' event occurs and is intended to close the websocket or perform other cleanup actions.

    Return
    ------
    AsyncIterator[dict]
        An asynchronous iterator yielding Twilio websocket messages.

    Notes
    -----
    - The function does not stop yielding messages after the 'stop' event; it continues
      to yield any remaining messages in the iterator after performing the closure action.

    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if item["event"] == "stop":
                logger.info("Stop message received, closing websocket")
                await close_func()
            yield item


async def twilio_media_to_audio_bytes_step(
    async_iter: AsyncIterator[dict],
) -> AsyncIterator[bytes]:
    """
    Data flow step that extracts the audio bytes from Twilio media stream messages.

    Twilio media stream messages come in as JSON, which are usually converted to Python dictionaries.
    Each dictionary contains a 'media' key with a 'payload' sub-key. The function decodes the base64-encoded payload into bytes
    representing audio data and yields these bytes.

    Parameters
    ----------
    async_iter : AsyncIterator[dict]
        An asynchronous iterator over Twilio media stream messages.

    Yields
    ------
    AsyncIterator[bytes]
        An asynchronous iterator yielding audio data in bytes.

    Raises
    ------
    ValueError
        If a message in the stream does not contain a 'media' key or if the 'media' key
        does not have a 'payload' sub-key.

    Notes
    -----
    - The function is specifically designed to handle media streams from Twilio,
      converting the audio into bytes objects for further processing.
    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if "media" not in item:
                raise ValueError(f"Message {item} has no media")
            yield base64.b64decode(item["media"]["payload"])


async def audio_bytes_to_twilio_media_step(
    async_iter: AsyncIterator[bytes], stream_sid: AwaitableOrObj[str]
) -> AsyncIterator[dict]:
    """
    Data flow step that formats audio bytes into outgoing Twilio media stream messages.

    To play audio over a Twilio media stream, messages must be formatted as JSON with the audio bytes base64-encoded.
    This step converts raw audio data into the correct format for Twilio.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator over audio bytes
    stream_sid: AwaitableOrObj[str]
        The Twilio stream id for the outgoing audio.  This is passed with the :class:`~voice_stream.CallStart` message when the call is initiated.

    Returns
    -------
    AsyncIterator[bytes]
        An asynchronous iterator yielding outgoing Twilio media stream messages.

    """
    resolved_stream_sid = None
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for packet in owned_aiter:
            encoded_data = base64.b64encode(packet).decode("utf-8")
            if not resolved_stream_sid:
                resolved_stream_sid = await resolve_awaitable_or_obj(stream_sid)
            yield {
                "event": "media",
                "streamSid": resolved_stream_sid,
                "media": {"payload": encoded_data},
            }


async def twilio_format_events_step(
    async_iter: AsyncIterator[Dict],
) -> AsyncIterator[BaseEvent]:
    """
    Data flow step to formats Twilio event messages into VoiceStream call status objects.

    This function converts Twilio event messages into the VoiceStream generic call event classes.  These include
    :class:`~voice_stream.events.CallStarted`, :class:`~voice_stream.events.CallEnded`, and :class:`~voice_stream.events.AnsweringMachineDetection`

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator over Twilio event messages.

    Yields
    ------
    AsyncIterator[BaseEvent]
        An asynchronous iterator yielding `BaseEvent` objects that represent formatted
        Twilio events.

    Notes
    -----
    - Twilio messages that don't correspond to events are ignored.
    - For 'start' events, it creates a `CallStarted` event object with the call and
      stream IDs.
    - For 'stop' events, it creates a `CallEnded` event object.
    """

    def format_events(input):
        if input["event"] == "start":
            return CallStarted(
                call_id=input["start"]["callSid"], stream_id=input["start"]["streamSid"]
            )
        if input["event"] == "stop":
            return CallEnded()

    return map_step(async_iter, format_events)


@dataclasses.dataclass
class TwilioInputFlow:
    """Experimental"""

    # Stream of audio bytes.
    audio: AsyncIterator[bytes]
    # All the events coming from the main stream in twilio, excluding raw audio.
    events: AsyncIterator[BaseEvent]
    all_twilio_messages: Optional[AsyncIterator[dict]]
    call_sid_f: asyncio.Future[str]
    stream_sid_f: asyncio.Future[str]
    inbound_queue_f: asyncio.Future[asyncio.Queue[BaseEvent]]
    outbound_queue_f: asyncio.Future[asyncio.Queue[BaseEvent]]
    current_calls: Dict[str, Any]

    @classmethod
    def create(
        cls,
        source: AsyncIterator[dict],
        close_func: Callable[[], None],
        current_calls: Dict[str, Any],
        expose_all_messages: bool = False,
    ) -> TwilioInputFlow:
        """
        Creates a basic flow to receive audio and call event data from Twilio.
        expose_all_messages - If true, all incoming twilio messages, including raw audio, will be set through the 'all_twilio_messages' iterator.  Otherwise this iterator will be empty.
        """
        stream = map_str_to_json_step(source)
        stream = filter_step(stream, lambda x: x["event"] != "connected")
        if expose_all_messages:
            stream, all_twilio_messages = fork_step(stream)
        else:
            all_twilio_messages = None
        stream = twilio_check_sequence_step(stream)
        # Retrieve values once the call is connected.
        stream, call_sid_f = extract_value_step(
            stream, value=lambda x: x["start"]["callSid"]
        )
        stream, stream_sid_f = extract_value_step(
            stream, value=lambda x: x["streamSid"]
        )
        outbound_queue_f = map_future(call_sid_f, lambda x: current_calls[x].outbound)
        inbound_queue_f = map_future(call_sid_f, lambda x: current_calls[x].inbound)
        audio, events = partition_step(stream, lambda x: x["event"] == "media")
        audio = twilio_media_to_audio_bytes_step(audio)
        events = twilio_close_on_stop_step(events, close_func=close_func)
        events = twilio_format_events_step(events)
        events = merge_step(events, queue_source(inbound_queue_f))
        return cls(
            audio,
            events,
            all_twilio_messages,
            call_sid_f,
            stream_sid_f,
            inbound_queue_f,
            outbound_queue_f,
            current_calls,
        )



from typing import Any, List

from langchain_community.chat_message_histories import ChatMessageHistory
from langchain_core.messages import BaseMessage
from langchain_core.prompts import MessagesPlaceholder
from langchain_core.runnables.history import GetSessionHistoryCallable


class InMemoryChatMessageHistories(GetSessionHistoryCallable):
    def __init__(self):
        self.histories = {}

    def __call__(self, session_id):
        return self.histories.setdefault(session_id, ChatMessageHistory())


class LastKMessagesPlaceholder(MessagesPlaceholder):
    """Variant of Langchain MessagesPlaceholder that only includes the last X messages."""

    variable_name: str
    """Name of variable to use as messages."""

    optional: bool = False

    max_messages: int
    """Max messages to include.  Use 0 for no limit."""

    def __init__(
        self,
        variable_name: str,
        max_messages: int,
        *,
        optional: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            variable_name=variable_name,
            optional=optional,
            max_messages=max_messages,
            **kwargs,
        )

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format messages from kwargs.

        Args:
            **kwargs: Keyword arguments to use for formatting.

        Returns:
            List of BaseMessage.
        """
        value = (
            kwargs.get(self.variable_name, [])
            if self.optional
            else kwargs[self.variable_name]
        )
        if not isinstance(value, list):
            raise ValueError(
                f"variable {self.variable_name} should be a list of base messages, "
                f"got {value}"
            )
        for v in value:
            if not isinstance(v, BaseMessage):
                raise ValueError(
                    f"variable {self.variable_name} should be a list of base messages,"
                    f" got {value}"
                )
        if self.max_messages > 0:
            value = value[-self.max_messages :]
        return value
